from . import ast
from .ast import AST_BY_TOKEN_TYPE, Paragraph
from .spec import (
    TokenSpecification,
    FORMAT_TOKEN_TYPES,
    TERMINAL_TOKEN_TYPES,
    NONFORMAT_TOKEN_TYPES,
    QUOTE_TOKEN_TYPES,
    EOF,
)


class Parser:
    def __init__(self, tokens):
        self.tokens = tokens
        self.token_iter = iter(self.tokens)
        self.eof = self.tokens[-1]
        self._stack = []
        self._tree = []

    @property
    def tree(self):
        return self._tree

    def print(self):
        for node in self.tree:
            print(node.eval(), end="")

    def parse(self):
        self._tree = []
        elems = []
        format_tokens = []
        current_token = next(self.token_iter)
        create_new_paragraph = False
        is_quote = False
        quote_token = None
        is_code_block = False

        while current_token != self.eof:
            format_token = None
            node = None

            if current_token.type in QUOTE_TOKEN_TYPES:
                is_quote = True
                quote_token = current_token
            elif current_token.type == TokenSpecification.CODE_BLOCK.name:
                is_code_block = True

            if current_token.type in FORMAT_TOKEN_TYPES:
                format_tokens.append(current_token)
            else:
                if current_token.type != TokenSpecification.NEWLINE.name:
                    node = AST_BY_TOKEN_TYPE[TokenSpecification.TEXT.name](
                        current_token.value
                    )
                    elems.append(node)

            current_token = next(self.token_iter)

            while format_tokens and not create_new_paragraph:
                if (
                    is_quote
                    and quote_token.type == TokenSpecification.INLINE_QUOTE.name
                    and current_token.type == TokenSpecification.NEWLINE.name
                ) or (
                    is_quote
                    and quote_token.type == TokenSpecification.BLOCK_QUOTE.name
                    and current_token.type == EOF
                ):
                    format_tokens.pop()
                    node = AST_BY_TOKEN_TYPE[quote_token.type](
                        node, md_tag=quote_token.value
                    )
                    is_quote = False
                    create_new_paragraph = True
                    quote_token = None
                    elems.append(node)
                elif (
                    is_quote
                    and quote_token.type == TokenSpecification.BLOCK_QUOTE.name
                    and current_token.type in NONFORMAT_TOKEN_TYPES
                ):
                    text_value = current_token.value
                    if node is not None:
                        text_value = node.value + text_value
                    node = AST_BY_TOKEN_TYPE[TokenSpecification.TEXT.name](text_value)
                elif current_token.type in FORMAT_TOKEN_TYPES:
                    if current_token.type == format_tokens[-1].type:
                        format_token = format_tokens.pop()

                        if format_token.type == TokenSpecification.BOLD_ITALIC.name:
                            node = ast.BoldText(ast.ItalicText(node))
                        else:
                            node = AST_BY_TOKEN_TYPE[format_token.type](
                                node, md_tag=format_token.value
                            )

                        if not format_tokens:
                            elems.append(node)

                        if is_code_block:
                            is_code_block = False
                    else:
                        format_tokens.append(current_token)
                elif not is_code_block and current_token.type in TERMINAL_TOKEN_TYPES:
                    create_new_paragraph = True
                elif is_code_block:
                    text_value = current_token.value
                    if node is not None:
                        text_value = node.value + text_value
                    node = AST_BY_TOKEN_TYPE[TokenSpecification.TEXT.name](text_value)
                else:
                    text_value = current_token.value
                    node = AST_BY_TOKEN_TYPE[TokenSpecification.TEXT.name](text_value)

                if current_token != self.eof:
                    current_token = next(self.token_iter)

            if create_new_paragraph or current_token.type in TERMINAL_TOKEN_TYPES:
                self._tree.append(Paragraph(elems))
                elems = []
                create_new_paragraph = False


class ParseError(Exception):
    pass



import re

from .spec import TokenSpecification, Token, CODE_TOKEN_TYPES, EOF


def tokenize(code, skip_newline=False):
    line_count = code.count("\n") + 1
    eof = Token(EOF, value="", line=line_count, column=len(code))

    token_iter = tokenize_generator(code, skip_newline)
    current_token = next(token_iter, eof)
    code_token = None
    tokens = []
    code_tokens = []
    text_tokens = []

    while current_token != eof:
        if (
            current_token.line == eof.line
            and current_token.column == eof.column
            and current_token.value == ""
        ):
            break

        if current_token.type in CODE_TOKEN_TYPES:
            if code_tokens and current_token.type == code_tokens[-1].type:
                code_tokens.pop()
                code_token = None
            else:
                code_token = current_token
                code_tokens.append(current_token)
                tokens.append(code_token)

            current_token = next(token_iter, eof)

        while current_token != eof and (
            current_token.type == "TEXT"
            or current_token.type == "SPACE"
            or code_token is not None
        ):
            text_tokens.append(current_token)
            current_token = next(token_iter, eof)
            if code_token is not None and current_token.type == code_token.type:
                code_token = None

        if text_tokens:
            concat_text = [t.value for t in text_tokens]
            concat_text_token = Token(
                "TEXT",
                value="".join(concat_text),
                line=text_tokens[0].line,
                column=text_tokens[0].column,
            )
            text_tokens = []
            tokens.append(concat_text_token)

        if current_token != eof:
            tokens.append(current_token)

        current_token = next(token_iter, eof)

    tokens.append(eof)

    return tokens


def tokenize_generator(code, skip_newline=False):
    tok_regex = "|".join(
        "(?P<%s>%s)%s" % token_spec.value for token_spec in list(TokenSpecification)
    )
    line_num = 1
    line_start = 0
    for mo in re.finditer(tok_regex, code):
        token_type = mo.lastgroup
        value = mo.group()
        column = mo.start() - line_start
        if token_type == "NEWLINE":
            line_start = mo.end()
            line_num += 1
            if skip_newline:
                continue
        yield Token(token_type, value, line_num, column)



from .lexer import tokenize
from .parser import Parser


class Compiler:
    def __init__(self, text):
        self._text = text
        self._tokens = tokenize(self._text)
        self._parser = Parser(self._tokens)
        self._code = ""

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, text):
        self._text = text
        self._tokens = tokenize(self._text)

    @property
    def tokens(self):
        return self._tokens

    @property
    def code(self):
        return self._code

    def compile(self, markdown=False):
        print(self._tokens)
        if not self._parser.tree:
            self._parser.parse()
        self._code = ""
        for node in self._parser.tree:
            self._code = self._code + node.eval(markdown=markdown)
        self._code = self._code.strip()
        return self._code



from .spec import TokenSpecification


class Paragraph:
    HTML_TAG = "p"

    def __init__(self, elements):
        self.elements = elements

    def open(self, markdown=False):
        if markdown:
            return ""
        else:
            return f"<{self.HTML_TAG}>"

    def close(self, markdown=False):
        if markdown:
            return "\n"
        else:
            return f"</{self.HTML_TAG}>"

    def eval(self, markdown=False):
        evaluated = "".join([elem.eval(markdown) for elem in self.elements])
        return f"{self.open(markdown)}{evaluated}{self.close(markdown)}"


class Text:
    HTML_TAG = None
    MD_TAG = None

    def __init__(self, value, md_tag=None, style=None):
        self.value = value
        self.md_tag = None
        self.style = style

    def eval(self, markdown=False):
        return self.value


class FormattedText(Text):
    HTML_TAG = ""
    MD_TAG = ""
    HAS_CLOSE_MD_TAG = True

    def __init__(self, value, md_tag=None, style=None):
        super().__init__(value, style=style)
        if md_tag is None:
            md_tag = self.MD_TAG
        self.md_tag = md_tag

    def open(self, markdown=False):
        if markdown:
            open_tag = self.md_tag
        else:
            if self.style:
                open_tag = f"<{self.HTML_TAG} style='{self.style}'>"
            else:
                open_tag = f"<{self.HTML_TAG}>"
        return open_tag

    def close(self, markdown=False):
        if markdown:
            close_tag = self.md_tag if self.HAS_CLOSE_MD_TAG else ""
        else:
            close_tag = f"</{self.HTML_TAG}>"
        return close_tag

    def eval(self, markdown=False):
        return f"{self.open(markdown)}{self.value.eval(markdown)}{self.close(markdown)}"


class ParagraphText(FormattedText):
    HTML_TAG = "p"
    MD_TAG = "\n"


class BoldText(FormattedText):
    HTML_TAG = "b"
    MD_TAG = "**"


class ItalicText(FormattedText):
    HTML_TAG = "i"
    MD_TAG = "*"


class UnderlineText(FormattedText):
    HTML_TAG = "u"
    MD_TAG = "__"


class StrikethroughText(FormattedText):
    HTML_TAG = "s"
    MD_TAG = "~~"


class InlineCode(FormattedText):
    HTML_TAG = "code"
    MD_TAG = "`"


class CodeBlock(FormattedText):
    HTML_TAG = "code"
    MD_TAG = "```"

    def close(self, markdown=False):
        if markdown:
            close_tag = self.MD_TAG
        else:
            close_tag = f"</{self.HTML_TAG}>"
        return close_tag

    def eval(self, markdown=False):
        evaluated = (
            f"{self.open(markdown)}{self.value.eval(markdown)}{self.close(markdown)}"
        )
        if markdown:
            return evaluated
        else:
            return f"<pre>{evaluated}</pre>"


class InlineQuote(FormattedText):
    HTML_TAG = "q"
    MD_TAG = "> "
    HAS_CLOSE_MD_TAG = False


class BlockQuote(FormattedText):
    HTML_TAG = "blockquote"
    MD_TAG = ">>> "
    HAS_CLOSE_MD_TAG = False


class SpoilerText(FormattedText):
    HTML_TAG = "span"
    MD_TAG = "||"

    def __init__(self, value, md_tag=None, style="color: black; background: black;"):
        super().__init__(value, md_tag=md_tag, style=style)


AST_BY_TOKEN_TYPE = {
    TokenSpecification.TEXT.name: Text,
    TokenSpecification.BOLD.name: BoldText,
    TokenSpecification.ITALIC.name: ItalicText,
    TokenSpecification.UNDERLINE.name: UnderlineText,
    TokenSpecification.STRIKETHROUGH.name: StrikethroughText,
    TokenSpecification.BLOCK_QUOTE.name: BlockQuote,
    TokenSpecification.INLINE_QUOTE.name: InlineQuote,
    TokenSpecification.CODE_BLOCK.name: CodeBlock,
    TokenSpecification.INLINE_CODE.name: InlineCode,
    TokenSpecification.SPOILER.name: SpoilerText,
}



import collections
import enum


EOF = "EOF"

Token = collections.namedtuple("Token", ["type", "value", "line", "column"])
TokenSpec = collections.namedtuple("TokenSpec", ["name", "pattern", "instances"])


class TokenSpecification(enum.Enum):
    CODE_BLOCK = ("CODE_BLOCK", r"`", r"{3}\w*")
    INLINE_CODE = ("INLINE_CODE", r"`", r"{1}")
    SPOILER = ("SPOILER", r"\|", r"{2}")
    BOLD_ITALIC = ("BOLD_ITALIC", r"\*", r"{3}")
    BOLD = ("BOLD", r"\*", r"{2}")
    UNDERLINE = ("UNDERLINE", r"_", r"{2}")
    ITALIC = ("ITALIC", r"\*|_", r"{1}")
    STRIKETHROUGH = ("STRIKETHROUGH", r"~", r"{2}")
    BLOCK_QUOTE = ("BLOCK_QUOTE", r">", r"{3}")
    INLINE_QUOTE = ("INLINE_QUOTE", r">", r"{1}")
    NEWLINE = ("NEWLINE", r"\n", "")
    SPACE = ("SPACE", r"[ \t]+", "")
    TEXT = ("TEXT", r"[\S\s]?", "")


FORMAT_TOKEN_TYPES = [
    TokenSpecification.CODE_BLOCK.name,
    TokenSpecification.BLOCK_QUOTE.name,
    TokenSpecification.INLINE_CODE.name,
    TokenSpecification.INLINE_QUOTE.name,
    TokenSpecification.BOLD.name,
    TokenSpecification.BOLD_ITALIC.name,
    TokenSpecification.SPOILER.name,
    TokenSpecification.UNDERLINE.name,
    TokenSpecification.ITALIC.name,
    TokenSpecification.STRIKETHROUGH.name,
]


NONFORMAT_TOKEN_TYPES = [
    TokenSpecification.TEXT.name,
    TokenSpecification.SPACE.name,
    TokenSpecification.NEWLINE.name,
    EOF,
]


TERMINAL_TOKEN_TYPES = [
    TokenSpecification.NEWLINE.name,
    EOF,
]


QUOTE_TOKEN_TYPES = [
    TokenSpecification.INLINE_QUOTE.name,
    TokenSpecification.BLOCK_QUOTE.name,
]


CODE_TOKEN_TYPES = [
    TokenSpecification.CODE_BLOCK.name,
    TokenSpecification.INLINE_CODE.name,
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['discord_markdown']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'discord-markdown',
    'version': '0.3.2',
    'description': 'Convert discord messages to HTML',
    'long_description': '# Discord Markdown\n\nI needed to render some Discord chat logs as HTML, and found that the Markdown implementation in Discord isn\'t quite compliant with Common Markdown as Discord uses a simplified version. \n\nSo I wrote this library that allows you to convert a discord message written in the Markdown formatting syntax specified [here](https://support.discordapp.com/hc/en-us/articles/210298617-Markdown-Text-101-Chat-Formatting-Bold-Italic-Underline-) to html.\n\n\n## Installation\n\nYou can install the library from `pypi`:\n\n```sh\npip install discord-markdown\n```\n\n## Usage\n\n```python\nfrom discord_markdown.discord_markdown import convert_to_html\n\ntext = "_This_ **is** an __example__.\\nThis should be a different paragraph."\nhtml = convert_to_html(text)\n\nassert html == \'<p><i>This</i> <b>is</b> an <u>example</u>.</p><p>This should be a different paragraph.</p>\'\n```\n',
    'author': 'BitJockey',
    'author_email': 'bitjockey@jackpoint.network',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/bitjockey42/discord-markdown',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from .compiler import Compiler
from .parser import Parser
from .lexer import tokenize


def convert_to_html(text):
    return Compiler(text).compile(markdown=False)


def parse_text(text):
    tokens = tokenize(text)
    parser = Parser(tokens)
    parser.parse()
    return parser.tree


class DiscordMarkdown:
    def __init__(self, text):
        self._compiler = Compiler(text)

    def markdown(self):
        return self._compiler.compile(True)

    def html(self):
        return self._compiler.compile(False)
